# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Tuple

import jax
import jax.numpy as jnp
from flax import nnx
from jax.sharding import Mesh
from transformers import Qwen3Config
from vllm.config import VllmConfig

from tpu_inference.distributed.jax_parallel_state import get_pp_group
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.layers.jax import JaxModule
from tpu_inference.layers.jax.embed import JaxEmbed
from tpu_inference.layers.jax.linear import JaxEinsum, JaxLinear
from tpu_inference.layers.jax.moe.moe import JaxMoE
from tpu_inference.layers.jax.moe.utils import (get_expert_parallelism,
                                                select_moe_backend)
from tpu_inference.layers.jax.norm import JaxRmsNorm
from tpu_inference.layers.jax.pp_utils import (PPMissingLayer,
                                               get_start_end_layer)
from tpu_inference.layers.jax.quantization.configs import QuantizationConfig
from tpu_inference.logger import init_logger
from tpu_inference.models.jax.jax_intermediate_tensor import \
    JaxIntermediateTensors
from tpu_inference.models.jax.qwen3 import Qwen3Attention
from tpu_inference.models.jax.utils.weight_utils import LoadableWithIterator

logger = init_logger(__name__)

init_fn = nnx.initializers.uniform()


class Qwen3MoeSparseMoeBlock(JaxModule):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 prefix: str = ""):
        config = vllm_config.model_config.hf_text_config
        dtype = vllm_config.model_config.dtype
        quant_config = vllm_config.quant_config

        # --- Sharding Config ---
        edf_sharding = (None, None, None)
        expert_axis_name = edf_sharding[0]
        num_expert_parallelism = get_expert_parallelism(expert_axis_name, mesh)
        use_ep = num_expert_parallelism > 1
        moe_backend = select_moe_backend(use_ep)

        # Router
        self.gate = JaxLinear(
            config.hidden_size,
            config.num_experts,
            rngs=rng,
            use_bias=False,
            quant_config=quant_config,
            prefix=prefix + ".gate",
        )
        self.gate.num_experts_per_tok = config.num_experts_per_tok

        # Shared Expert
        shared_expert_intermediate_size = getattr(
            config, "shared_expert_intermediate_size", 0)
        if shared_expert_intermediate_size > 0:
            raise NotImplementedError(
                f"Shared expert is not implemented yet. Found {shared_expert_intermediate_size=} in config."
            )
        else:
            self.shared_expert = None

        # Experts (Routed)
        self.experts = JaxMoE(
            dtype=dtype,
            num_local_experts=config.num_experts,
            hidden_size=config.hidden_size,
            intermediate_size_moe=config.moe_intermediate_size,
            hidden_act=config.hidden_act,
            rngs=rng,
            router=self.gate,
            mesh=mesh,
            activation_ffw_td=(ShardingAxisName.MLP_DATA, None),
            activation_ffw_ted=(ShardingAxisName.MLP_DATA, None, None),
            edf_sharding=(None, ),
            efd_sharding=(None, ),
            apply_expert_weight_before_computation=False,
            expert_axis_name=expert_axis_name,
            num_expert_parallelism=num_expert_parallelism,
            moe_backend=moe_backend,
            quant_config=quant_config,
            prefix=prefix + ".experts")

    def __call__(self, x: jax.Array) -> jax.Array:
        out = self.experts(x)
        if self.shared_expert is not None:
            out += self.shared_expert(x)
        return out


class Qwen3MoeDecoderLayer(JaxModule):

    def __init__(self,
                 config: Qwen3Config,
                 dtype: jnp.dtype,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 kv_cache_dtype: str,
                 quant_config: QuantizationConfig,
                 layer_idx: int,
                 vllm_config: VllmConfig,
                 prefix: str = ""):
        rms_norm_eps = config.rms_norm_eps
        hidden_size = config.hidden_size

        self.input_layernorm = JaxRmsNorm(
            hidden_size,
            epsilon=rms_norm_eps,
            param_dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".input_layernorm",
        )
        self.self_attn = Qwen3Attention(
            config=config,
            dtype=dtype,
            rng=rng,
            mesh=mesh,
            kv_cache_dtype=kv_cache_dtype,
            quant_config=quant_config,
            prefix=prefix + ".self_attn",
        )
        self.post_attention_layernorm = JaxRmsNorm(
            hidden_size,
            epsilon=rms_norm_eps,
            param_dtype=dtype,
            scale_init=nnx.with_partitioning(init_fn, (None, )),
            rngs=rng,
            quant_config=quant_config,
            prefix=prefix + ".post_attention_layernorm",
        )

        mlp_only_layers = getattr(config, "mlp_only_layers", [])
        if (layer_idx not in mlp_only_layers) and (
                config.num_experts > 0 and
            (layer_idx + 1) % config.decoder_sparse_step == 0):
            self.mlp = Qwen3MoeSparseMoeBlock(vllm_config=vllm_config,
                                              rng=rng,
                                              mesh=mesh,
                                              prefix=prefix + ".mlp")
        else:
            raise NotImplementedError(
                f"Non-sparse MLP is not implemented yet. Found {mlp_only_layers=}, {config.num_experts=}, and {config.decoder_sparse_step=} in config."
            )

    def __call__(
        self,
        kv_cache: jax.Array,
        x: jax.Array,
        attention_metadata: AttentionMetadata,
    ) -> Tuple[jax.Array, jax.Array]:
        hidden_states = self.input_layernorm(x)
        kv_cache, attn_output = self.self_attn(
            kv_cache,
            hidden_states,
            attention_metadata,
        )
        attn_output += x

        residual = attn_output
        attn_output = self.post_attention_layernorm(attn_output)
        outputs = self.mlp(attn_output)
        outputs = residual + outputs
        return kv_cache, outputs


class Qwen3MoeModel(JaxModule):

    def __init__(self,
                 vllm_config: VllmConfig,
                 rng: nnx.Rngs,
                 mesh: Mesh,
                 prefix: str = "") -> None:
        model_config = vllm_config.model_config
        hf_config = model_config.hf_config
        vocab_size = model_config.get_vocab_size()
        dtype = model_config.dtype
        rms_norm_eps = hf_config.rms_norm_eps
        hidden_size = hf_config.hidden_size

        self.is_first_rank = get_pp_group().is_first_rank
        self.is_last_rank = get_pp_group().is_last_rank

        if self.is_first_rank or (hf_config.tie_word_embeddings
                                  and self.is_last_rank):
            self.embed_tokens = JaxEmbed(
                num_embeddings=vocab_size,
                features=hidden_size,
                param_dtype=dtype,
                embedding_init=nnx.with_partitioning(init_fn, ("model", None)),
                rngs=rng,
                quant_config=vllm_config.quant_config,
                prefix=prefix + ".embed_tokens",
            )
        else:
            self.embed_tokens = PPMissingLayer()

        self.start_layer, self.end_layer = get_start_end_layer(
            hf_config.num_hidden_layers,
            get_pp_group().rank_in_group,
            get_pp_group().world_size)

        self.layers = []
        for i in range(hf_config.num_hidden_layers):
            if self.start_layer <= i < self.end_layer:
                self.layers.append(
                    Qwen3MoeDecoderLayer(
                        config=hf_config,
                        dtype=dtype,
                        rng=rng,
                        mesh=mesh,
                        kv_cache_dtype=vllm_config.cache_config.cache_dtype,
                        quant_config=vllm_config.quant_config,
                        layer_idx=i,
                        vllm_config=vllm_config,
                        prefix=f"{prefix}.layers.{i}",
                    ))
            else:
                self.layers.append(PPMissingLayer())

        if self.is_last_rank:
            self.norm = JaxRmsNorm(
                hidden_size,
                epsilon=rms_norm_eps,
                param_dtype=dtype,
                scale_init=nnx.with_partitioning(init_fn, (None, )),
                rngs=rng,
                quant_config=vllm_config.quant_config,
                prefix=prefix + ".final_layernorm",
            )
        else:
            self.norm = PPMissingLayer()

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        inputs_embeds: Optional[jax.Array] = None,
    ) -> Tuple[List[jax.Array], jax.Array]:
        if self.is_first_rank:
            assert inputs_embeds is None
            inputs_embeds = self.embed_tokens(input_ids)
        else:
            assert inputs_embeds is not None

        x = inputs_embeds
        new_kv_caches = []
        for i, layer in enumerate(self.layers):
            if isinstance(layer, PPMissingLayer):
                new_kv_caches.append(kv_caches[i])
                continue
            kv_cache = kv_caches[i]
            kv_cache, x = layer(kv_cache, x, attention_metadata)
            new_kv_caches.append(kv_cache)

        if self.is_last_rank:
            x = self.norm(x)

        return new_kv_caches, x


class Qwen3MoeForCausalLM(JaxModule, LoadableWithIterator):

    def __init__(self, vllm_config: VllmConfig, rng_key: jax.Array,
                 mesh: Mesh) -> None:
        self.vllm_config = vllm_config
        rng = nnx.Rngs(rng_key)
        self.mesh = mesh

        self.model = Qwen3MoeModel(
            vllm_config=vllm_config,
            rng=rng,
            mesh=mesh,
            prefix="model",
        )
        model_config = vllm_config.model_config
        if not model_config.hf_config.tie_word_embeddings:
            if self.model.is_last_rank:
                vocab_size = model_config.get_vocab_size()
                hidden_size = model_config.hf_config.hidden_size
                self.lm_head = JaxEinsum(
                    einsum_str="TD,DV->TV",
                    kernel_shape=(hidden_size, vocab_size),
                    dtype=model_config.dtype,
                    rngs=rng,
                    quant_config=vllm_config.quant_config,
                    prefix="lm_head",
                )
            else:
                self.lm_head = PPMissingLayer()

    def __call__(
        self,
        kv_caches: List[jax.Array],
        input_ids: jax.Array,
        attention_metadata: AttentionMetadata,
        inputs_embeds: Optional[jax.Array] = None,
        _input_positions=None,
        _layer_name_to_kv_cache=None,
        _lora_metadata=None,
        intermediate_tensors: JaxIntermediateTensors | None = None,
        is_first_rank: bool = True,
        is_last_rank: bool = True,
        *args,
    ) -> Tuple[List[jax.Array], jax.Array | JaxIntermediateTensors,
               List[jax.Array]]:
        if not is_first_rank:
            assert intermediate_tensors is not None
            inputs_embeds = intermediate_tensors["hidden_states"]
        kv_caches, x = self.model(
            kv_caches,
            input_ids,
            attention_metadata,
            inputs_embeds,
        )
        if not is_last_rank:
            x = JaxIntermediateTensors(tensors={"hidden_states": x}, )
        return kv_caches, x, []

    def compute_logits(self, hidden_states: jax.Array) -> jax.Array:
        if hasattr(self, 'lm_head'):
            return self.lm_head(hidden_states)

        assert isinstance(self.model.embed_tokens, JaxEmbed)
        return self.model.embed_tokens.decode(hidden_states)



# SPDX-License-Identifier: Apache-2.0
"""Reduce-scatter matmul kernel module."""

from .kernel import get_kernel_name, reduce_scatter_matmul

__all__ = [
    "reduce_scatter_matmul",
    "get_kernel_name",
]



# SPDX-License-Identifier: Apache-2.0
"""Bidirectional Reduce-Scatter Matmul with M-Split Algorithm.

This implementation uses BOTH left and right neighbors simultaneously to double
the effective communication bandwidth. The key insight is to split the M dimension
into N blocks (one per device), and each block into TOP and BOT halves:
  - LEFT direction handles all TOP halves
  - RIGHT direction handles all BOT halves

This avoids the collision problem where both directions would compute for the
same shard at the midpoint step.

Setup (8 devices example):
- Each device has: x[M, K_shard], y[N, K_shard] where K is sharded
- M is split into 8 blocks, each block split into TOP and BOT halves
- Output: each device gets its M_block (TOP + BOT) with full K reduction
- D0 owns Block 0 (rows 0 to M/8-1)
- D1 owns Block 1 (rows M/8 to 2M/8-1)
- etc.

Bidirectional ring:
- LEFT direction: D0 → D7 → D6 → D5 → D4 → D3 → D2 → D1 → D0 (send to left)
- RIGHT direction: D0 → D1 → D2 → D3 → D4 → D5 → D6 → D7 → D0 (send to right)

M-SPLIT BIDIRECTIONAL ALGORITHM:
================================

Split M into N blocks, each block into TOP (first half) and BOT (second half):
  - LEFT handles: B0_TOP, B1_TOP, B2_TOP, ..., B(N-1)_TOP
  - RIGHT handles: B0_BOT, B1_BOT, B2_BOT, ..., B(N-1)_BOT

Each direction does COMPLETE reduce-scatter (N-1 steps) for its halves.

From Device 0's perspective (8 devices):

  Step 0 (Prologue):
    Barrier with both neighbors
    LEFT:  Compute P₀(B1_TOP) for block 1's top half
    RIGHT: Compute P₀(B7_BOT) for block 7's bot half
    Send both to neighbors

  Steps 1 to N-2: For each step s:
    Signal neighbors, wait for capacity
    Start bidirectional DMA:
      - LEFT: send to D7, receive from D1
      - RIGHT: send to D1, receive from D7
    DELAYED COMPUTATION (overlapped with DMA):
      - LEFT:  Compute P₀(B(s+1)_TOP) for target block's top half
      - RIGHT: Compute P₀(B(7-s)_BOT) for target block's bot half
    Wait for DMAs
    Accumulate computation results to received data

  Final Step:
    Final DMA exchange
    Compute own block contributions:
      - LEFT:  Compute P₀(B0_TOP) for own block's top half
      - RIGHT: Compute P₀(B0_BOT) for own block's bot half
    Accumulate to received data
    Write output:
      - TOP half from LEFT direction
      - BOT half from RIGHT direction

KEY INSIGHTS:
1. NO COLLISION: LEFT always computes TOP halves, RIGHT always computes BOT halves
2. Even at midpoint (step 3 for 8 devices), they compute DIFFERENT halves of same block
3. PERFECTLY BALANCED: Every step has exactly 2 half-block matmuls
4. NO IDLE STEPS: Both directions always have compute work
5. 2X BANDWIDTH: Both ICI directions fully utilized
6. GOOD OVERLAP: Compute overlaps with bidirectional DMA
"""

import functools
from typing import Any, NamedTuple

import jax
import jax.numpy as jnp
from jax import lax
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu

P = jax.sharding.PartitionSpec
Ref = Any


def mod(x: jax.Array, n: int) -> jax.Array:
    """Modulo operation that works with JAX arrays."""
    return lax.rem(x + n, n)


class KernelConfig(NamedTuple):
    """Configuration for the kernel."""

    num_devices: int
    m_block: int  # M rows per device = M // num_devices
    m_half_block: int  # Half of m_block = m_block // 2
    # Tile sizes for matmul
    bm: int = 128  # Block size for M dimension
    bn: int = 128  # Block size for N dimension
    bk: int = 128  # Block size for K dimension


def tiled_matmul_hbm(
    x_hbm_ref: Ref,  # [M, K_shard] in HBM (full buffer)
    y_hbm_ref: Ref,  # [N, K_shard] in HBM
    out_hbm_ref: Ref,  # [M_half_block, N] in HBM
    # VMEM scratch buffers
    x_vmem_ref: Ref,  # [bm, bk] in VMEM
    y_vmem_ref: Ref,  # [bn, bk] in VMEM
    acc_vmem_ref: Ref,  # [bm, bn] in VMEM (float32 for accumulation)
    out_vmem_ref: Ref,  # [bm, bn] in VMEM
    # Semaphores
    copy_sem: Ref,  # DMA semaphore
    *,
    m_block_idx: int
    | jax.Array,  # Which M block (multiply by bm to get offset)
    m_size: int,  # Number of M rows to process
    bm: int,
    bn: int,
    bk: int,
):
    """Tiled matmul: out = x[m_block_idx*bm:m_block_idx*bm+m_size, :] @ y.T.

    This function tiles the matmul computation using async_copy for HBM<->VMEM transfers.
    The result OVERWRITES the output buffer (does not accumulate to existing values).

    IMPORTANT: Uses block indices (multiplied by bm) instead of raw offsets to help
    the Mosaic compiler prove tile alignment at compile time.

    Args:
        x_hbm_ref: Full X buffer in HBM [M, K_shard]
        y_hbm_ref: Y buffer in HBM [N, K_shard]
        out_hbm_ref: Output buffer in HBM [m_size, N]
        x_vmem_ref: VMEM scratch for x tile [bm, bk]
        y_vmem_ref: VMEM scratch for y tile [bn, bk]
        acc_vmem_ref: VMEM scratch for accumulator [bm, bn] in float32
        out_vmem_ref: VMEM scratch for output tile [bm, bn]
        copy_sem: DMA semaphore for async copies
        m_block_idx: Starting block index (offset = m_block_idx * bm)
        m_size: Number of rows to process
        bm: Block size for M dimension
        bn: Block size for N dimension
        bk: Block size for K dimension
    """
    _, k_shard = x_hbm_ref.shape
    n_total, _ = y_hbm_ref.shape

    num_m_tiles = m_size // bm
    num_n_tiles = n_total // bn
    num_k_tiles = k_shard // bk

    for m_tile in range(num_m_tiles):
        # Compute global m index using block index arithmetic
        # This form helps the compiler prove alignment since it's (block_idx + tile) * bm
        global_m_tile = m_block_idx + m_tile
        for n_tile in range(num_n_tiles):
            n_start = n_tile * bn

            # Zero the accumulator
            acc_vmem_ref[...] = jnp.zeros((bm, bn), dtype=jnp.float32)

            # Iterate over K tiles
            for k_tile in range(num_k_tiles):
                k_start = k_tile * bk

                # Copy x tile from HBM to VMEM
                # Use global_m_tile * bm for the offset (compiler can prove alignment)
                x_copy = pltpu.make_async_copy(
                    src_ref=x_hbm_ref.at[pl.ds(global_m_tile * bm, bm),
                                         pl.ds(k_start, bk)],
                    dst_ref=x_vmem_ref,
                    sem=copy_sem,
                )
                x_copy.start()
                x_copy.wait()

                # Copy y tile from HBM to VMEM
                y_copy = pltpu.make_async_copy(
                    src_ref=y_hbm_ref.at[pl.ds(n_start, bn),
                                         pl.ds(k_start, bk)],
                    dst_ref=y_vmem_ref,
                    sem=copy_sem,
                )
                y_copy.start()
                y_copy.wait()

                # Compute partial result and accumulate
                x_f32 = x_vmem_ref[...].astype(jnp.float32)
                y_f32 = y_vmem_ref[...].astype(jnp.float32)
                acc_vmem_ref[...] = acc_vmem_ref[...] + jnp.dot(x_f32, y_f32.T)

            # Convert accumulator to output dtype and store
            out_vmem_ref[...] = acc_vmem_ref[...].astype(out_hbm_ref.dtype)

            # Copy result from VMEM to HBM
            # Use m_tile * bm for output offset (within the output buffer)
            out_copy = pltpu.make_async_copy(
                src_ref=out_vmem_ref,
                dst_ref=out_hbm_ref.at[pl.ds(m_tile * bm, bm),
                                       pl.ds(n_start, bn)],
                sem=copy_sem,
            )
            out_copy.start()
            out_copy.wait()


def tiled_add_hbm(
    src_hbm_ref: Ref,  # [M_half_block, N] in HBM
    dst_hbm_ref: Ref,  # [M_half_block, N] in HBM (will be modified in place)
    # VMEM scratch buffers
    src_vmem_ref: Ref,  # [bm, bn] in VMEM
    dst_vmem_ref: Ref,  # [bm, bn] in VMEM
    # Semaphores
    copy_sem: Ref,  # DMA semaphore
    *,
    bm: int,
    bn: int,
):
    """Tiled addition: dst += src, with HBM inputs.

    Adds the source buffer to the destination buffer in-place using async_copy.
    """
    m_size, n_total = src_hbm_ref.shape

    num_m_tiles = m_size // bm
    num_n_tiles = n_total // bn

    for m_tile in range(num_m_tiles):
        m_start = m_tile * bm
        for n_tile in range(num_n_tiles):
            n_start = n_tile * bn

            # Copy src tile from HBM to VMEM
            src_copy = pltpu.make_async_copy(
                src_ref=src_hbm_ref.at[pl.ds(m_start, bm),
                                       pl.ds(n_start, bn)],
                dst_ref=src_vmem_ref,
                sem=copy_sem,
            )
            src_copy.start()
            src_copy.wait()

            # Copy dst tile from HBM to VMEM
            dst_copy = pltpu.make_async_copy(
                src_ref=dst_hbm_ref.at[pl.ds(m_start, bm),
                                       pl.ds(n_start, bn)],
                dst_ref=dst_vmem_ref,
                sem=copy_sem,
            )
            dst_copy.start()
            dst_copy.wait()

            # Add and store back to dst_vmem
            result = src_vmem_ref[...].astype(
                jnp.float32) + dst_vmem_ref[...].astype(jnp.float32)
            dst_vmem_ref[...] = result.astype(dst_hbm_ref.dtype)

            # Copy result from VMEM to HBM
            out_copy = pltpu.make_async_copy(
                src_ref=dst_vmem_ref,
                dst_ref=dst_hbm_ref.at[pl.ds(m_start, bm),
                                       pl.ds(n_start, bn)],
                sem=copy_sem,
            )
            out_copy.start()
            out_copy.wait()


def _kernel(
    # Inputs (in HBM)
    x_ref: Ref,  # [M, K_shard]
    y_ref: Ref,  # [N, K_shard]
    # Output (in HBM)
    out_ref: Ref,  # [M_block, N] - this device's output block
    # Scratch space for communication (in HBM)
    # 4 slots: [0,1] for LEFT direction, [2,3] for RIGHT direction
    scratch_ref: Ref,  # [4, M_half_block, N]
    # Scratch space for computation results (in HBM)
    computation_scratch_ref: Ref,  # [2, M_half_block, N]
    # VMEM scratch buffers for tiled matmul
    x_vmem_ref: Ref,  # [bm, bk] in VMEM
    y_vmem_ref: Ref,  # [bn, bk] in VMEM
    acc_vmem_ref: Ref,  # [bm, bn] in VMEM
    out_vmem_ref: Ref,  # [bm, bn] in VMEM
    # VMEM scratch for tiled add
    add_vmem_ref: Ref,  # [bm, bn] in VMEM for tiled_add src
    # Semaphores
    send_sem: Ref,  # DMA semaphore for sending
    recv_sem: Ref,  # DMA semaphore for receiving
    copy_sem: Ref,  # DMA semaphore for local copies
    left_capacity_sem: Ref,  # REGULAR semaphore for flow control (left)
    right_capacity_sem: Ref,  # REGULAR semaphore for flow control (right)
    *,
    config: KernelConfig,
    axis_name: str,
):
    """Bidirectional Reduce-Scatter Matmul Kernel with M-split algorithm.

    Grid: (num_devices,) where each iteration is one ring step.

    Key insight: Split M into N blocks, each block into TOP and BOT halves.
    - LEFT direction handles all TOP halves (reduced via left ring)
    - RIGHT direction handles all BOT halves (reduced via right ring)

    This ensures:
    - No collision at midpoint (different halves)
    - Perfect load balance (every step has 2 half-block matmuls)
    - Full bandwidth utilization (both directions active)
    """
    num_devices = config.num_devices
    m_block = config.m_block
    m_half_block = config.m_half_block
    bm, bn, bk = config.bm, config.bn, config.bk

    # Grid index = ring step
    ring_step = pl.program_id(0)

    # Device topology
    my_id = lax.axis_index(axis_name)
    left_neighbor = mod(my_id - 1, num_devices)
    right_neighbor = mod(my_id + 1, num_devices)

    # Double buffer management for bidirectional communication
    # Left direction: slots 0 and 1
    # Right direction: slots 2 and 3
    left_working_slot = lax.rem(ring_step, 2)  # 0 or 1
    left_receiving_slot = 1 - left_working_slot
    right_working_slot = 2 + lax.rem(ring_step, 2)  # 2 or 3
    right_receiving_slot = 5 - right_working_slot  # 3 or 2

    # Computation scratch slots (double buffered)
    left_compute_slot = 0
    right_compute_slot = 1

    m_total, k_shard = x_ref.shape
    n_total, _ = y_ref.shape

    num_steps = num_devices
    is_first_step = ring_step == 0
    is_last_step = ring_step == num_steps - 1

    # Semaphores are now passed as parameters from scratch_shapes
    # (send_sem, recv_sem, copy_sem, left_capacity_sem, right_capacity_sem)

    # =========================================================================
    # Helper functions
    # =========================================================================

    def get_left_target_block(step):
        """LEFT direction: compute TOP half of block (my_id + step + 1) % N."""
        return mod(my_id + step + 1, num_devices)

    def get_right_target_block(step):
        """RIGHT direction: compute BOT half of block (my_id - step - 1) % N."""
        return mod(my_id - step - 1, num_devices)

    def compute_matmul_top_half(block_idx, out_slot):
        """Compute matmul for TOP half of specified block.

        Result: x[block_top, :] @ y.T → computation_scratch_ref[out_slot]
        """
        # Convert to block index: m_start = block_idx * m_block
        # m_block_idx = m_start // bm = block_idx * (m_block // bm)
        m_block_idx = block_idx * (m_block // bm)

        tiled_matmul_hbm(
            x_hbm_ref=x_ref,
            y_hbm_ref=y_ref,
            out_hbm_ref=computation_scratch_ref.at[out_slot],
            x_vmem_ref=x_vmem_ref,
            y_vmem_ref=y_vmem_ref,
            acc_vmem_ref=acc_vmem_ref,
            out_vmem_ref=out_vmem_ref,
            copy_sem=copy_sem,
            m_block_idx=m_block_idx,
            m_size=m_half_block,
            bm=bm,
            bn=bn,
            bk=bk,
        )

    def compute_matmul_bot_half(block_idx, out_slot):
        """Compute matmul for BOT half of specified block.

        Result: x[block_bot, :] @ y.T → computation_scratch_ref[out_slot]
        """
        # Convert to block index: m_start = block_idx * m_block + m_half_block
        # m_block_idx = m_start // bm = block_idx * (m_block // bm) + (m_half_block // bm)
        m_block_idx = block_idx * (m_block // bm) + (m_half_block // bm)

        tiled_matmul_hbm(
            x_hbm_ref=x_ref,
            y_hbm_ref=y_ref,
            out_hbm_ref=computation_scratch_ref.at[out_slot],
            x_vmem_ref=x_vmem_ref,
            y_vmem_ref=y_vmem_ref,
            acc_vmem_ref=acc_vmem_ref,
            out_vmem_ref=out_vmem_ref,
            copy_sem=copy_sem,
            m_block_idx=m_block_idx,
            m_size=m_half_block,
            bm=bm,
            bn=bn,
            bk=bk,
        )

    def accumulate_computation_to_slot(compute_slot, dst_slot):
        """Add computation_scratch_ref[compute_slot] to scratch_ref[dst_slot]."""
        tiled_add_hbm(
            src_hbm_ref=computation_scratch_ref.at[compute_slot],
            dst_hbm_ref=scratch_ref.at[dst_slot],
            src_vmem_ref=add_vmem_ref,
            dst_vmem_ref=out_vmem_ref,
            copy_sem=copy_sem,
            bm=bm,
            bn=bn,
        )

    def copy_computation_to_slot(compute_slot, dst_slot):
        """Copy computation_scratch_ref[compute_slot] to scratch_ref[dst_slot]."""
        local_copy = pltpu.make_async_copy(
            src_ref=computation_scratch_ref.at[compute_slot],
            dst_ref=scratch_ref.at[dst_slot],
            sem=copy_sem,
        )
        local_copy.start()
        local_copy.wait()

    def local_barrier():
        """Barrier with both neighbors using double-barrier pattern."""
        barrier_sem = pltpu.get_barrier_semaphore()

        pltpu.semaphore_signal(
            barrier_sem,
            inc=1,
            device_id=(left_neighbor, ),
            device_id_type=pltpu.DeviceIdType.MESH,
        )
        pltpu.semaphore_signal(
            barrier_sem,
            inc=1,
            device_id=(right_neighbor, ),
            device_id_type=pltpu.DeviceIdType.MESH,
        )
        pltpu.semaphore_wait(barrier_sem, 2)

        @functools.partial(pl.run_scoped,
                           second_barrier=pltpu.SemaphoreType.REGULAR)
        def _(second_barrier):
            pltpu.semaphore_signal(
                second_barrier,
                inc=1,
                device_id=(left_neighbor, ),
                device_id_type=pltpu.DeviceIdType.MESH,
            )
            pltpu.semaphore_signal(
                second_barrier,
                inc=1,
                device_id=(right_neighbor, ),
                device_id_type=pltpu.DeviceIdType.MESH,
            )
            pltpu.semaphore_wait(second_barrier, 2)

    def signal_left_neighbor():
        """Signal left neighbor that we are ready to receive from them."""
        pltpu.semaphore_signal(
            left_capacity_sem,
            inc=1,
            device_id=(left_neighbor, ),
            device_id_type=pltpu.DeviceIdType.MESH,
        )

    def signal_right_neighbor():
        """Signal right neighbor that we are ready to receive from them."""
        pltpu.semaphore_signal(
            right_capacity_sem,
            inc=1,
            device_id=(right_neighbor, ),
            device_id_type=pltpu.DeviceIdType.MESH,
        )

    # Get target blocks for current step
    left_target_block = get_left_target_block(ring_step)
    right_target_block = get_right_target_block(ring_step)

    # =========================================================================
    # STEP 0: Prologue - Barrier, compute, prepare for sending
    # =========================================================================
    @pl.when(is_first_step)
    def _prologue():
        # Barrier with both neighbors
        local_barrier()

        # Compute TOP half of left target block (for LEFT direction)
        compute_matmul_top_half(left_target_block, left_compute_slot)
        # Compute BOT half of right target block (for RIGHT direction)
        compute_matmul_bot_half(right_target_block, right_compute_slot)

        # Copy computation results to scratch slots for sending
        # LEFT result goes to left_working_slot (will be sent left in step 1)
        # RIGHT result goes to right_working_slot (will be sent right in step 1)
        copy_computation_to_slot(left_compute_slot, left_working_slot)
        copy_computation_to_slot(right_compute_slot, right_working_slot)

    # =========================================================================
    # STEPS 1 to N-1: Main loop with bidirectional DMA and compute
    # =========================================================================
    @pl.when(~is_first_step)
    def _main_loop():
        # Signal both neighbors that we're ready to receive from them
        signal_left_neighbor()
        signal_right_neighbor()

        # Wait for both neighbors to be ready to receive from us
        pltpu.semaphore_wait(left_capacity_sem, 1)
        pltpu.semaphore_wait(right_capacity_sem, 1)

        # Bidirectional DMA:
        # - Send left_receiving_slot to left neighbor, receive from right into left_working_slot
        # - Send right_receiving_slot to right neighbor, receive from left into right_working_slot
        remote_copy_to_left = pltpu.make_async_remote_copy(
            src_ref=scratch_ref.at[left_receiving_slot],
            dst_ref=scratch_ref.at[left_working_slot],
            send_sem=send_sem,
            recv_sem=recv_sem,
            device_id=(left_neighbor, ),
            device_id_type=pltpu.DeviceIdType.MESH,
        )
        remote_copy_to_left.start()

        remote_copy_to_right = pltpu.make_async_remote_copy(
            src_ref=scratch_ref.at[right_receiving_slot],
            dst_ref=scratch_ref.at[right_working_slot],
            send_sem=send_sem,
            recv_sem=recv_sem,
            device_id=(right_neighbor, ),
            device_id_type=pltpu.DeviceIdType.MESH,
        )
        remote_copy_to_right.start()

        # DELAYED COMPUTATION (overlapped with DMA):
        # - LEFT: compute TOP half of current target block
        # - RIGHT: compute BOT half of current target block
        compute_matmul_top_half(left_target_block, left_compute_slot)
        compute_matmul_bot_half(right_target_block, right_compute_slot)

        # Wait for both DMAs to complete
        remote_copy_to_left.wait()
        remote_copy_to_right.wait()

        # At last step: write output
        @pl.when(is_last_step)
        def _epilogue():
            # Accumulate own contribution to received data
            accumulate_computation_to_slot(left_compute_slot,
                                           left_working_slot)
            accumulate_computation_to_slot(right_compute_slot,
                                           right_working_slot)

            # Write output:
            # - TOP half (from LEFT direction) → out_ref[0:m_half_block, :]
            # - BOT half (from RIGHT direction) → out_ref[m_half_block:m_block, :]
            top_copy = pltpu.make_async_copy(
                src_ref=scratch_ref.at[left_working_slot],
                dst_ref=out_ref.at[pl.ds(0, m_half_block), :],
                sem=copy_sem,
            )
            top_copy.start()
            top_copy.wait()

            bot_copy = pltpu.make_async_copy(
                src_ref=scratch_ref.at[right_working_slot],
                dst_ref=out_ref.at[pl.ds(m_half_block, m_half_block), :],
                sem=copy_sem,
            )
            bot_copy.start()
            bot_copy.wait()

        # For non-last steps: accumulate and continue
        @pl.when(~is_last_step)
        def _accumulate():
            accumulate_computation_to_slot(left_compute_slot,
                                           left_working_slot)
            accumulate_computation_to_slot(right_compute_slot,
                                           right_working_slot)


def bidirectional_reduce_scatter_matmul(
    x: jax.Array,  # [M, K_shard] - K is sharded
    y: jax.Array,  # [N, K_shard] - K is sharded
    *,
    axis_name: str = "x",
    bm: int = 128,
    bn: int = 128,
    bk: int = 128,
) -> jax.Array:
    """Bidirectional reduce-scatter matmul with M-split algorithm.

    Computes: reduce_scatter(x @ y.T, scatter_dim=0)

    Where:
    - x @ y.T is computed with K sharded across devices
    - Result is scattered on M dimension (each device gets M/num_devices rows)

    The algorithm splits M into N blocks (one per device), and each block into
    TOP and BOT halves. LEFT direction handles TOP halves, RIGHT handles BOT halves.
    This achieves:
    - 2x ICI bandwidth utilization
    - Perfect load balance (every step has compute work)
    - No collision (directions always process different halves)

    Args:
        x: Input tensor [M, K_shard] where K is sharded across devices
        y: Weight tensor [N, K_shard] where K is sharded across devices
        axis_name: Name of the device axis for collective operations
        bm: Block size for M dimension (must divide M_half_block)
        bn: Block size for N dimension (must divide N)
        bk: Block size for K dimension (must divide K_shard)

    Returns:
        Output tensor [M_block, N] where M is scattered across devices
    """
    num_devices = lax.psum(1, axis_name)

    m_total, k_shard = x.shape
    n_total, _ = y.shape

    assert (
        m_total % num_devices == 0
    ), f"M ({m_total}) must be divisible by num_devices ({num_devices})"

    m_block = m_total // num_devices

    assert m_block % 2 == 0, f"M_block ({m_block}) must be divisible by 2"

    m_half_block = m_block // 2

    # Validate tile sizes
    assert (m_half_block % bm == 0
            ), f"M_half_block ({m_half_block}) must be divisible by bm ({bm})"
    assert n_total % bn == 0, f"N ({n_total}) must be divisible by bn ({bn})"
    assert k_shard % bk == 0, f"K_shard ({k_shard}) must be divisible by bk ({bk})"

    config = KernelConfig(
        num_devices=num_devices,
        m_block=m_block,
        m_half_block=m_half_block,
        bm=bm,
        bn=bn,
        bk=bk,
    )

    # Output shape: [M_block, N]
    out_shape = jax.ShapeDtypeStruct((m_block, n_total), x.dtype)

    # Scratch shapes for communication (4 slots for bidirectional double buffering)
    scratch_shape = jax.ShapeDtypeStruct((4, m_half_block, n_total), x.dtype)

    # Scratch shapes for computation (2 slots)
    computation_scratch_shape = jax.ShapeDtypeStruct(
        (2, m_half_block, n_total), x.dtype)

    # VMEM scratch shapes
    x_vmem_shape = pltpu.VMEM((bm, bk), x.dtype)
    y_vmem_shape = pltpu.VMEM((bn, bk), y.dtype)
    acc_vmem_shape = pltpu.VMEM((bm, bn), jnp.float32)
    out_vmem_shape = pltpu.VMEM((bm, bn), x.dtype)
    add_vmem_shape = pltpu.VMEM((bm, bn), x.dtype)

    # Grid: one iteration per ring step
    grid = (num_devices, )

    def kernel_fn(x_ref, y_ref, out_ref, scratch_ref, computation_scratch_ref,
                  x_vmem_ref, y_vmem_ref, acc_vmem_ref, out_vmem_ref,
                  add_vmem_ref, send_sem, recv_sem, copy_sem,
                  left_capacity_sem, right_capacity_sem):
        _kernel(
            x_ref=x_ref,
            y_ref=y_ref,
            out_ref=out_ref,
            scratch_ref=scratch_ref,
            computation_scratch_ref=computation_scratch_ref,
            x_vmem_ref=x_vmem_ref,
            y_vmem_ref=y_vmem_ref,
            acc_vmem_ref=acc_vmem_ref,
            out_vmem_ref=out_vmem_ref,
            add_vmem_ref=add_vmem_ref,
            send_sem=send_sem,
            recv_sem=recv_sem,
            copy_sem=copy_sem,
            left_capacity_sem=left_capacity_sem,
            right_capacity_sem=right_capacity_sem,
            config=config,
            axis_name=axis_name,
        )

    out, _, _ = pl.pallas_call(
        kernel_fn,
        out_shape=(out_shape, scratch_shape, computation_scratch_shape),
        grid_spec=pltpu.PrefetchScalarGridSpec(
            num_scalar_prefetch=0,
            in_specs=[
                pl.BlockSpec(memory_space=pltpu.ANY),  # x
                pl.BlockSpec(memory_space=pltpu.ANY),  # y
            ],
            out_specs=[
                pl.BlockSpec(memory_space=pltpu.ANY),  # out
                pl.BlockSpec(memory_space=pltpu.ANY),  # scratch
                pl.BlockSpec(memory_space=pltpu.ANY),  # computation_scratch
            ],
            scratch_shapes=[
                x_vmem_shape,
                y_vmem_shape,
                acc_vmem_shape,
                out_vmem_shape,
                add_vmem_shape,
                # DMA semaphores
                pltpu.SemaphoreType.DMA,  # send_sem
                pltpu.SemaphoreType.DMA,  # recv_sem
                pltpu.SemaphoreType.DMA,  # copy_sem
                # Flow control semaphores
                pltpu.SemaphoreType.REGULAR,  # left_capacity_sem
                pltpu.SemaphoreType.REGULAR,  # right_capacity_sem
            ],
            grid=grid,
        ),
        compiler_params=pltpu.CompilerParams(
            collective_id=0,
            dimension_semantics=("arbitrary", ),
        ),
    )(x, y)

    return out


# Algorithm diagram for reference
ALGORITHM_DIAGRAM = """
BIDIRECTIONAL REDUCE-SCATTER MATMUL WITH M-SPLIT ALGORITHM
===========================================================

M dimension split into N blocks, each block split into TOP and BOT halves:

Full M dimension (8 devices):
┌─────────┬─────────┬─────────┬─────────┬─────────┬─────────┬─────────┬─────────┐
│ Block 0 │ Block 1 │ Block 2 │ Block 3 │ Block 4 │ Block 5 │ Block 6 │ Block 7 │
└─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴─────────┘

Each block split:
┌────┬────┬────┬────┬────┬────┬────┬────┬────┬────┬────┬────┬────┬────┬────┬────┐
│B0_T│B0_B│B1_T│B1_B│B2_T│B2_B│B3_T│B3_B│B4_T│B4_B│B5_T│B5_B│B6_T│B6_B│B7_T│B7_B│
└────┴────┴────┴────┴────┴────┴────┴────┴────┴────┴────┴────┴────┴────┴────┴────┘
  │         │         │         │         │         │         │         │
  └─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴─────────┘
                            LEFT (all TOP halves)
        └─────────┴─────────┴─────────┴─────────┴─────────┴─────────┴─────────┘
                            RIGHT (all BOT halves)

Direction assignment:
  LEFT:  All TOP halves → reduces via left ring (D0→D7→D6→...→D1→D0)
  RIGHT: All BOT halves → reduces via right ring (D0→D1→D2→...→D7→D0)

Device 0's computation schedule (8 devices):
┌──────────┬─────────────────────────┬─────────────────────────┐
│   Step   │     LEFT (TOP half)     │    RIGHT (BOT half)     │
├──────────┼─────────────────────────┼─────────────────────────┤
│    0     │ P₀(B1_TOP) → send to D7 │ P₀(B7_BOT) → send to D1 │
│    1     │ P₀(B2_TOP) + accum      │ P₀(B6_BOT) + accum      │
│    2     │ P₀(B3_TOP) + accum      │ P₀(B5_BOT) + accum      │
│    3     │ P₀(B4_TOP) + accum      │ P₀(B4_BOT) + accum      │  ← Same block, DIFFERENT halves!
│    4     │ P₀(B5_TOP) + accum      │ P₀(B3_BOT) + accum      │
│    5     │ P₀(B6_TOP) + accum      │ P₀(B2_BOT) + accum      │
│    6     │ P₀(B7_TOP) + accum      │ P₀(B1_BOT) + accum      │
│  Final   │ P₀(B0_TOP) → output TOP │ P₀(B0_BOT) → output BOT │
└──────────┴─────────────────────────┴─────────────────────────┘

Pipeline timeline (Device 0):
Time ────────────────────────────────────────────────────────────────────────────►

Step:     0      1      2      3      4      5      6     Final
         ───    ───    ───    ───    ───    ───    ───    ─────

LEFT:    ████   ████   ████   ████   ████   ████   ████   ████
(TOP)    B1_T   B2_T   B3_T   B4_T   B5_T   B6_T   B7_T   B0_T

RIGHT:   ████   ████   ████   ████   ████   ████   ████   ████
(BOT)    B7_B   B6_B   B5_B   B4_B   B3_B   B2_B   B1_B   B0_B

L-DMA:          ════   ════   ════   ════   ════   ════   ════
                →D7    →D7    →D7    →D7    →D7    →D7    →D7

R-DMA:          ════   ════   ════   ════   ════   ════   ════
                →D1    →D1    →D1    →D1    →D1    →D1    →D1

Matmuls: [2]    [2]    [2]    [2]    [2]    [2]    [2]    [2]

KEY BENEFITS:
✓ NO COLLISION: LEFT and RIGHT always process DIFFERENT halves
✓ PERFECT BALANCE: Every step has exactly 2 half-block matmuls
✓ NO IDLE STEPS: Both directions always have compute work
✓ 2X BANDWIDTH: Both ICI directions fully utilized
✓ GOOD OVERLAP: Compute overlaps with bidirectional DMA

Final output at Device 0:
┌─────────────────────────────────────────────────────────────┐
│  Block 0 TOP (from LEFT):  P₀ + P₁ + P₂ + ... + P₇        │
├─────────────────────────────────────────────────────────────┤
│  Block 0 BOT (from RIGHT): P₀ + P₁ + P₂ + ... + P₇        │
└─────────────────────────────────────────────────────────────┘
Combined: Complete Block 0 with full reduction from all 8 devices ✓
"""

if __name__ == "__main__":
    print(ALGORITHM_DIAGRAM)



# SPDX-License-Identifier: Apache-2.0
"""Test script for the bidirectional reduce-scatter matmul kernel.

This script tests the bidirectional reduce-scatter matmul kernel with M-split
algorithm for correctness by comparing its output with a reference implementation.

The kernel implements a bidirectional ring-based reduce-scatter fused with matmul where:
- Each device has x[M, K_shard] and y[N, K_shard] (K is sharded)
- Output is x @ y.T with M dimension scattered across devices
- Uses M-split algorithm: M split into N blocks, each block into TOP/BOT halves
- LEFT direction handles all TOP halves, RIGHT handles all BOT halves
- Uses delayed computation for compute-communication overlap

Usage:
    pytest tests/experimental/test_bidirection_reduce_scatter_matmul.py

Requirements:
    - Must be run on a TPU VM with JAX installed
    - Requires multiple TPU devices for multi-device tests
    - JAX with TPU support: pip install jax[tpu]
"""
import jax
import jax.numpy as jnp
import numpy as np
from absl.testing import absltest, parameterized
from experimental.reduce_scatter_matmul.bidirection_reduce_scatter_matmul import \
    bidirectional_reduce_scatter_matmul
from jax import lax, random
from jax._src import test_util as jtu
from jax.experimental import mesh_utils
from jax.experimental.shard_map import shard_map
from jax.sharding import Mesh, NamedSharding
from jax.sharding import PartitionSpec as P


def reduce_scatter_matmul_reference(
    x_shard: jax.Array,
    y_shard: jax.Array,
    axis_name: str = "x",
) -> jax.Array:
    """Reference implementation using lax.psum_scatter.

    Args:
        x_shard: Input tensor [M, K_shard] where K is sharded
        y_shard: Weight tensor [N, K_shard] where K is sharded
        axis_name: Name of the device axis

    Returns:
        Output tensor [M_shard, N] where M is scattered
    """
    # Local matmul: [M, K_shard] @ [K_shard, N] = [M, N]
    local_result = jnp.dot(x_shard, y_shard.T)

    # Reduce-scatter along M dimension
    # psum_scatter with tiled=True: scatter the result evenly across devices
    result = lax.psum_scatter(local_result,
                              axis_name,
                              scatter_dimension=0,
                              tiled=True)

    return result


def bidirectional_reduce_scatter_matmul_sharded(
    x: jax.Array,
    y: jax.Array,
    mesh: Mesh,
    axis_name: str = "x",
    bm: int = 128,
    bn: int = 128,
    bk: int = 128,
) -> jax.Array:
    """Wrapper that applies sharding to bidirectional_reduce_scatter_matmul.

    Args:
        x: Input tensor [M, K] with K sharded
        y: Weight tensor [N, K] with K sharded
        mesh: JAX mesh for device placement
        axis_name: Name of the device axis
        bm: Block size for M dimension
        bn: Block size for N dimension
        bk: Block size for K dimension

    Returns:
        Output tensor [M, N] with M sharded
    """

    @jax.jit
    def kernel_fn(x_shard, y_shard):
        return bidirectional_reduce_scatter_matmul(x_shard,
                                                   y_shard,
                                                   axis_name=axis_name,
                                                   bm=bm,
                                                   bn=bn,
                                                   bk=bk)

    sharded_fn = shard_map(
        kernel_fn,
        mesh=mesh,
        in_specs=(P(None, "x"), P(None, "x")),  # K sharded
        out_specs=P("x", None),  # M sharded
        check_rep=False,
    )

    return sharded_fn(x, y)


def _print_array_comparison(result, expected, name=""):
    """Print diagnostic info for array comparison."""
    result_np = np.asarray(result).astype(np.float32)
    expected_np = np.asarray(expected).astype(np.float32)

    diff = np.abs(result_np - expected_np)
    rel_diff = diff / (np.abs(expected_np) + 1e-8)

    print(f"\n  {name} Comparison:")
    print(
        f"    Result shape: {result_np.shape}, dtype: {np.asarray(result).dtype}"
    )
    print(
        f"    Expected shape: {expected_np.shape}, dtype: {np.asarray(expected).dtype}"
    )
    print(
        f"    Result range: [{float(result_np.min()):.4f}, {float(result_np.max()):.4f}]"
    )
    print(
        f"    Expected range: [{float(expected_np.min()):.4f}, {float(expected_np.max()):.4f}]"
    )
    print(f"    Max abs diff: {float(diff.max()):.6f}")
    print(f"    Max rel diff: {float(rel_diff.max()):.6f}")


class BidirectionalReduceScatterMatmulTest(parameterized.TestCase):
    """Parameterized tests for bidirectional reduce-scatter matmul kernel."""

    @parameterized.product(
        dtype=[jnp.bfloat16, jnp.float32],
        shape_multiplier=[1, 2, 4],  # Multiplier for base dimensions
        device_count=[2, 4, 8],
        use_random_values=[False, True],
    )
    def test_reduce_scatter_matmul(self, dtype, shape_multiplier, device_count,
                                   use_random_values):
        """Parameterized test covering dtypes, shapes, and device counts."""
        num_devices = jax.device_count()

        if device_count > num_devices or num_devices % device_count != 0:
            self.skipTest(
                f"Requires {device_count} devices, but only {num_devices} available or not divisible"
            )

        # Clear caches to avoid conflicts between different mesh configurations
        jax.clear_caches()

        # Base dimensions scaled by multiplier
        # For bidirectional kernel: M_block = 256 * multiplier, M_half_block = 128 * multiplier
        # Must be divisible by bm=128
        m = 256 * device_count * shape_multiplier
        k = 128 * device_count * shape_multiplier
        n = 128 * shape_multiplier

        # Create inputs
        if use_random_values:
            key = random.key(42)
            key1, key2 = random.split(key)
            x = random.uniform(key1, shape=(m, k),
                               dtype=jnp.float32).astype(dtype)
            y = random.uniform(key2, shape=(n, k),
                               dtype=jnp.float32).astype(dtype)
        else:
            x = jnp.ones((m, k), dtype=dtype)
            y = jnp.ones((n, k), dtype=dtype)

        # Setup mesh with subset of devices
        all_devices = jax.devices()
        selected_devices = np.array(all_devices[:device_count])
        mesh = Mesh(selected_devices, axis_names=("x", ))

        x_sharding = NamedSharding(mesh, P(None, "x"))
        y_sharding = NamedSharding(mesh, P(None, "x"))

        x_sharded = jax.device_put(x, x_sharding)
        y_sharded = jax.device_put(y, y_sharding)

        print(
            f"\n  Test: dtype={dtype}, m={m}, k={k}, n={n}, devices={device_count}, random={use_random_values}"
        )

        # Run kernel
        result = bidirectional_reduce_scatter_matmul_sharded(x_sharded,
                                                             y_sharded,
                                                             mesh,
                                                             axis_name="x")

        # Reference implementation
        def reference_fn(x_shard, y_shard):
            return reduce_scatter_matmul_reference(x_shard,
                                                   y_shard,
                                                   axis_name="x")

        reference_fn_sharded = shard_map(
            reference_fn,
            mesh=mesh,
            in_specs=(P(None, "x"), P(None, "x")),
            out_specs=P("x", None),
            check_rep=False,
        )

        expected = reference_fn_sharded(x_sharded, y_sharded)

        # Verify shape
        self.assertEqual(result.shape, expected.shape)

        # Verify dtype
        self.assertEqual(result.dtype, dtype)

        # Verify values with appropriate tolerance
        rtol = 0.05 if dtype == jnp.bfloat16 else 1e-4
        atol = 0.1 if dtype == jnp.bfloat16 else 1e-5
        np.testing.assert_allclose(result, expected, rtol=rtol, atol=atol)

    def test_output_shape(self):
        """Test that output shape is correct."""
        num_devices = jax.device_count()

        if num_devices < 2:
            self.skipTest("Requires at least 2 devices")

        m = 256 * num_devices
        k = 256 * num_devices
        n = 128
        m_per_device = m // num_devices

        x = jnp.ones((m, k), dtype=jnp.bfloat16)
        y = jnp.ones((n, k), dtype=jnp.bfloat16)

        devices = mesh_utils.create_device_mesh((num_devices, ))
        mesh = Mesh(devices, axis_names=("x", ))

        x_sharding = NamedSharding(mesh, P(None, "x"))
        y_sharding = NamedSharding(mesh, P(None, "x"))

        x_sharded = jax.device_put(x, x_sharding)
        y_sharded = jax.device_put(y, y_sharding)

        result = bidirectional_reduce_scatter_matmul_sharded(x_sharded,
                                                             y_sharded,
                                                             mesh,
                                                             axis_name="x")

        # Global shape should be [M, N] with M sharded
        self.assertEqual(result.shape, (m, n))

        # Check per-device local shape
        local_shape = result.addressable_shards[0].data.shape
        expected_local_shape = (m_per_device, n)
        self.assertEqual(local_shape, expected_local_shape)

    def test_slice_mapping(self):
        """Verify that each device outputs the correct M_block."""
        num_devices = jax.device_count()

        if num_devices < 2:
            self.skipTest("Requires at least 2 devices")

        m = 256 * num_devices
        k = 256 * num_devices
        n = 128
        m_per_device = m // num_devices

        # Create x where each m_block has a unique value pattern
        x = jnp.zeros((m, k), dtype=jnp.bfloat16)
        for i in range(num_devices):
            start = i * m_per_device
            end = (i + 1) * m_per_device
            x = x.at[start:end, :].set(float(i + 1))

        y = jnp.ones((n, k), dtype=jnp.bfloat16)

        devices = mesh_utils.create_device_mesh((num_devices, ))
        mesh = Mesh(devices, axis_names=("x", ))

        x_sharding = NamedSharding(mesh, P(None, "x"))
        y_sharding = NamedSharding(mesh, P(None, "x"))

        x_sharded = jax.device_put(x, x_sharding)
        y_sharded = jax.device_put(y, y_sharding)

        result = bidirectional_reduce_scatter_matmul_sharded(x_sharded,
                                                             y_sharded,
                                                             mesh,
                                                             axis_name="x")

        for device_idx, shard in enumerate(result.addressable_shards):
            # Check both TOP and BOT halves
            top_value = float(shard.data[0, 0])
            bot_value = float(shard.data[m_per_device // 2, 0])
            expected_value = (device_idx + 1) * k

            self.assertAlmostEqual(
                top_value,
                expected_value,
                delta=expected_value * 0.05,
                msg=f"TOP slice mapping wrong on device {device_idx}")
            self.assertAlmostEqual(
                bot_value,
                expected_value,
                delta=expected_value * 0.05,
                msg=f"BOT slice mapping wrong on device {device_idx}")

    def test_top_bot_consistency(self):
        """Test that TOP and BOT halves are consistent with reference."""
        num_devices = jax.device_count()

        if num_devices < 2:
            self.skipTest("Requires at least 2 devices")

        m = 256 * num_devices
        k = 256 * num_devices
        n = 128
        m_per_device = m // num_devices
        m_half = m_per_device // 2

        key = random.key(999)
        key1, key2 = random.split(key)

        x = random.uniform(key1, shape=(m, k),
                           dtype=jnp.float32).astype(jnp.bfloat16)
        y = random.uniform(key2, shape=(n, k),
                           dtype=jnp.float32).astype(jnp.bfloat16)

        devices = mesh_utils.create_device_mesh((num_devices, ))
        mesh = Mesh(devices, axis_names=("x", ))

        x_sharding = NamedSharding(mesh, P(None, "x"))
        y_sharding = NamedSharding(mesh, P(None, "x"))

        x_sharded = jax.device_put(x, x_sharding)
        y_sharded = jax.device_put(y, y_sharding)

        result = bidirectional_reduce_scatter_matmul_sharded(x_sharded,
                                                             y_sharded,
                                                             mesh,
                                                             axis_name="x")

        # Reference implementation
        def reference_fn(x_shard, y_shard):
            return reduce_scatter_matmul_reference(x_shard,
                                                   y_shard,
                                                   axis_name="x")

        reference_fn_sharded = shard_map(
            reference_fn,
            mesh=mesh,
            in_specs=(P(None, "x"), P(None, "x")),
            out_specs=P("x", None),
            check_rep=False,
        )

        expected = reference_fn_sharded(x_sharded, y_sharded)

        # Check each device's TOP and BOT halves match reference
        for device_idx, (result_shard, expected_shard) in enumerate(
                zip(result.addressable_shards, expected.addressable_shards)):
            result_data = np.asarray(result_shard.data).astype(np.float32)
            expected_data = np.asarray(expected_shard.data).astype(np.float32)

            top_result = result_data[:m_half, :]
            bot_result = result_data[m_half:, :]
            top_expected = expected_data[:m_half, :]
            bot_expected = expected_data[m_half:, :]

            np.testing.assert_allclose(
                top_result,
                top_expected,
                rtol=0.05,
                atol=0.1,
                err_msg=f"TOP half mismatch on device {device_idx}")
            np.testing.assert_allclose(
                bot_result,
                bot_expected,
                rtol=0.05,
                atol=0.1,
                err_msg=f"BOT half mismatch on device {device_idx}")


if __name__ == "__main__":
    absltest.main(testLoader=jtu.JaxTestLoader())
